@@ -15,12 +15,12 @@ from .exceptions import NoMatchingDataError, PaginationError
 from .mappings import Area, NEIGHBOURS, lookup_area
 from .misc import year_blocks, day_blocks
 from .parsers import parse_prices, parse_loads, parse_generation, \
-    parse_generation_per_plant, parse_installed_capacity_per_plant, \
-    parse_crossborder_flows, parse_unavailabilities, \
-    parse_contracted_reserve, parse_imbalance_prices_zip
+    parse_installed_capacity_per_plant, parse_crossborder_flows, \
+    parse_unavailabilities, parse_contracted_reserve, parse_imbalance_prices_zip, \
+    parse_netpositions
 
 __title__ = "entsoe-py"
-__version__ = "0.3.1"
+__version__ = "0.3.4"
 __author__ = "EnergieID.be"
 __license__ = "MIT"
 
@@ -183,6 +183,30 @@ class EntsoeRawClient:
         }
         response = self._base_request(params=params, start=start, end=end)
         return response.text
+    
+    def query_net_position_dayahead(self, country_code: Union[Area, str],
+                            start: pd.Timestamp, end: pd.Timestamp) -> str:
+        """
+        Parameters
+        ----------
+        country_code : Area|str
+        start : pd.Timestamp
+        end : pd.Timestamp
+
+        Returns
+        -------
+        str
+        """
+        area = lookup_area(country_code)
+        params = {
+            'documentType': 'A25',  # Allocation result document
+            'businessType': 'B09',  # net position
+            'Contract_MarketAgreement.Type': 'A01',  # daily
+            'in_Domain': area.code,
+            'out_Domain': area.code
+        }
+        response = self._base_request(params=params, start=start, end=end)
+        return response.text
 
     def query_load(self, country_code: Union[Area, str], start: pd.Timestamp,
                    end: pd.Timestamp) -> str:
@@ -518,15 +542,40 @@ class EntsoeRawClient:
             country_code_from=country_code_from,
             country_code_to=country_code_to, start=start, end=end,
             doctype="A61", contract_marketagreement_type="A04")
+    
+    def query_intraday_offered_capacity(
+        self, country_code_from: Union[Area, str],
+            country_code_to: Union[Area, str], start: pd.Timestamp,
+            end: pd.Timestamp, implicit:bool = True,**kwargs) -> str:
+        """
+        Parameters
+        ----------
+        country_code_from : Area|str
+        country_code_to : Area|str
+        start : pd.Timestamp
+        end : pd.Timestamp
+        implicit: bool (True = implicit - default for most borders. False = explicit - for instance BE-GB)
+
+        Returns
+        -------
+        str
+        """
+        return self._query_crossborder(
+            country_code_from=country_code_from,
+            country_code_to=country_code_to, start=start, end=end,
+            doctype="A31", contract_marketagreement_type="A07",
+            auction_type=("A01" if implicit==True else "A02"))
+
 
     def _query_crossborder(
             self, country_code_from: Union[Area, str],
             country_code_to: Union[Area, str], start: pd.Timestamp,
             end: pd.Timestamp, doctype: str,
-            contract_marketagreement_type: Optional[str] = None) -> str:
+            contract_marketagreement_type: Optional[str] = None,
+            auction_type: Optional[str] = None) -> str:
         """
-        Generic function called by query_crossborder_flows and
-        query_scheduled_exchanges.
+        Generic function called by query_crossborder_flows, 
+        query_scheduled_exchanges, query_net_transfer_capacity_DA/WA/MA/YA and query_.
 
         Parameters
         ----------
@@ -552,6 +601,9 @@ class EntsoeRawClient:
         if contract_marketagreement_type is not None:
             params[
                 'contract_MarketAgreement.Type'] = contract_marketagreement_type
+        if auction_type is not None:
+            params[
+                'Auction.Type'] = auction_type
 
         response = self._base_request(params=params, start=start, end=end)
         return response.text
@@ -582,6 +634,37 @@ class EntsoeRawClient:
         response = self._base_request(params=params, start=start, end=end)
         return response.content
 
+    def query_activated_balancing_energy(
+            self, country_code: Union[Area, str], start: pd.Timestamp,
+            end: pd.Timestamp, business_type: str, 
+            psr_type: Optional[str] = None) -> bytes:
+        """
+        Activated Balancing Energy [17.1.E]
+        Parameters
+        ----------
+        country_code : Area|str
+        start : pd.Timestamp
+        end : pd.Timestamp
+        business_type : str
+            type of contract (see mappings.BSNTYPE)
+        psr_type : str
+            filter query for a specific psr type
+
+        Returns
+        -------
+        bytes
+        """
+        area = lookup_area(country_code)
+        params = {
+            'documentType': 'A83',
+            'controlArea_Domain': area.code,
+            'businessType': business_type
+        }
+        if psr_type:
+            params.update({'psrType': psr_type})
+        response = self._base_request(params=params, start=start, end=end)
+        return response.content
+
     def query_contracted_reserve_prices(
             self, country_code: Union[Area, str], start: pd.Timestamp,
             end: pd.Timestamp, type_marketagreement_type: str,
@@ -796,6 +879,7 @@ class EntsoeRawClient:
             country_code=country_code, start=start, end=end,
             doctype="A80", docstatus='A13')
         return content
+    
 
 
 def paginated(func):
@@ -870,6 +954,29 @@ def day_limited(func):
 
 
 class EntsoePandasClient(EntsoeRawClient):
+    @year_limited
+    def query_net_position_dayahead(self, country_code: Union[Area, str],
+                            start: pd.Timestamp, end: pd.Timestamp) -> pd.Series:
+        """
+
+        Parameters
+        ----------
+        country_code
+        start
+        end
+
+        Returns
+        -------
+
+        """
+        area = lookup_area(country_code)
+        text = super(EntsoePandasClient, self).query_net_position_dayahead(
+            country_code=area, start=start, end=end)
+        series = parse_netpositions(text)
+        series = series.tz_convert(area.tz)
+        series = series.truncate(before=start, after=end)
+        return series
+
     @year_limited
     def query_day_ahead_prices(
             self, country_code: Union[Area, str], start: pd.Timestamp,
@@ -942,7 +1049,8 @@ class EntsoePandasClient(EntsoeRawClient):
     @year_limited
     def query_generation_forecast(
             self, country_code: Union[Area, str], start: pd.Timestamp,
-            end: pd.Timestamp, process_type: str = 'A01') -> pd.Series:
+            end: pd.Timestamp, process_type: str = 'A01',
+            nett: bool = False) -> Union[pd.DataFrame, pd.Series]:
         """
         Parameters
         ----------
@@ -950,18 +1058,20 @@ class EntsoePandasClient(EntsoeRawClient):
         start : pd.Timestamp
         end : pd.Timestamp
         process_type : str
+        nett : bool
+            condense generation and consumption into a nett number
 
         Returns
         -------
-        pd.Series
+        pd.DataFrame | pd.Series
         """
         area = lookup_area(country_code)
         text = super(EntsoePandasClient, self).query_generation_forecast(
             country_code=area, start=start, end=end, process_type=process_type)
-        series = parse_loads(text)
-        series = series.tz_convert(area.tz)
-        series = series.truncate(before=start, after=end)
-        return series
+        df = parse_generation(text, nett=nett)
+        df = df.tz_convert(area.tz)
+        df = df.truncate(before=start, after=end)
+        return df
 
     @year_limited
     def query_wind_and_solar_forecast(
@@ -986,7 +1096,7 @@ class EntsoePandasClient(EntsoeRawClient):
         text = super(EntsoePandasClient, self).query_wind_and_solar_forecast(
             country_code=area, start=start, end=end, psr_type=psr_type,
             process_type=process_type)
-        df = parse_generation(text)
+        df = parse_generation(text, nett=True)
         df = df.tz_convert(area.tz)
         df = df.truncate(before=start, after=end)
         return df
@@ -995,7 +1105,7 @@ class EntsoePandasClient(EntsoeRawClient):
     def query_generation(
             self, country_code: Union[Area, str], start: pd.Timestamp,
             end: pd.Timestamp, psr_type: Optional[str] = None,
-            **kwargs) -> pd.DataFrame:
+            nett: bool = False, **kwargs) -> pd.DataFrame:
         """
         Parameters
         ----------
@@ -1004,6 +1114,8 @@ class EntsoePandasClient(EntsoeRawClient):
         end : pd.Timestamp
         psr_type : str
             filter on a single psr type
+        nett : bool
+            condense generation and consumption into a nett number
 
         Returns
         -------
@@ -1012,7 +1124,7 @@ class EntsoePandasClient(EntsoeRawClient):
         area = lookup_area(country_code)
         text = super(EntsoePandasClient, self).query_generation(
             country_code=area, start=start, end=end, psr_type=psr_type)
-        df = parse_generation(text)
+        df = parse_generation(text, nett=nett)
         df = df.tz_convert(area.tz)
         df = df.truncate(before=start, after=end)
         return df
@@ -1223,7 +1335,7 @@ class EntsoePandasClient(EntsoeRawClient):
         ts = ts.tz_convert(area_from.tz)
         ts = ts.truncate(before=start, after=end)
         return ts
-
+    
     @year_limited
     def query_net_transfer_capacity_yearahead(
             self, country_code_from: Union[Area, str],
@@ -1255,6 +1367,38 @@ class EntsoePandasClient(EntsoeRawClient):
         ts = ts.truncate(before=start, after=end)
         return ts
 
+    @year_limited
+    def query_intraday_offered_capacity(
+        self, country_code_from: Union[Area, str],
+            country_code_to: Union[Area, str], start: pd.Timestamp,
+            end: pd.Timestamp, implicit:bool = True, **kwargs) -> pd.Series:
+        """
+        Note: Result will be in the timezone of the origin country  --> to check
+
+        Parameters
+        ----------
+        country_code_from : Area|str
+        country_code_to : Area|str
+        start : pd.Timestamp
+        end : pd.Timestamp
+        implicit: bool (True = implicit - default for most borders. False = explicit - for instance BE-GB)
+        Returns
+        -------
+        pd.Series
+        """
+        area_to = lookup_area(country_code_to)
+        area_from = lookup_area(country_code_from)
+        text = super(EntsoePandasClient, self).query_intraday_offered_capacity(
+            country_code_from=area_from,
+            country_code_to=area_to,
+            start=start,
+            end=end,
+            implicit=implicit)
+        ts = parse_crossborder_flows(text)
+        ts = ts.tz_convert(area_from.tz)
+        ts = ts.truncate(before=start, after=end)
+        return ts
+    
     @year_limited
     def query_imbalance_prices(
             self, country_code: Union[Area, str], start: pd.Timestamp,
@@ -1280,6 +1424,36 @@ class EntsoePandasClient(EntsoeRawClient):
         df = df.truncate(before=start, after=end)
         return df
 
+    @year_limited
+    def query_activated_balancing_energy(
+            self, country_code: Union[Area, str], start: pd.Timestamp,
+            end: pd.Timestamp, business_type: str, 
+            psr_type: Optional[str] = None) -> pd.DataFrame:
+        """
+        Activated Balancing Energy [17.1.E]
+        Parameters
+        ----------
+        country_code : Area|str
+        start : pd.Timestamp
+        end : pd.Timestamp
+        business_type: str
+            type of contract (see mappings.BSNTYPE)
+        psr_type : str
+            filter query for a specific psr type
+
+        Returns
+        -------
+        pd.DataFrame
+        """
+        area = lookup_area(country_code)
+        text = super(EntsoePandasClient, self).query_activated_balancing_energy(
+            country_code=area, start=start, end=end, 
+            business_type=business_type, psr_type=psr_type)
+        df = parse_contracted_reserve(text, area.tz, "quantity")
+        df = df.tz_convert(area.tz)
+        df = df.truncate(before=start, after=end)
+        return df
+    
     @year_limited
     @paginated
     def query_contracted_reserve_prices(
@@ -1485,7 +1659,7 @@ class EntsoePandasClient(EntsoeRawClient):
     def query_generation_per_plant(
             self, country_code: Union[Area, str], start: pd.Timestamp,
             end: pd.Timestamp, psr_type: Optional[str] = None,
-            **kwargs) -> pd.DataFrame:
+            nett: bool = False, **kwargs) -> pd.DataFrame:
         """
         Parameters
         ----------
@@ -1494,6 +1668,8 @@ class EntsoePandasClient(EntsoeRawClient):
         end : pd.Timestamp
         psr_type : str
             filter on a single psr type
+        nett : bool
+            condense generation and consumption into a nett number
 
         Returns
         -------
@@ -1502,7 +1678,8 @@ class EntsoePandasClient(EntsoeRawClient):
         area = lookup_area(country_code)
         text = super(EntsoePandasClient, self).query_generation_per_plant(
             country_code=area, start=start, end=end, psr_type=psr_type)
-        df = parse_generation_per_plant(text)
+        df = parse_generation(text, per_plant=True)
+        df.columns = df.columns.set_levels(df.columns.levels[0].str.encode('latin-1').str.decode('utf-8'), level=0)
         df = df.tz_convert(area.tz)
         # Truncation will fail if data is not sorted along the index in rare
         # cases. Ensure the dataframe is sorted:
@@ -1552,3 +1729,4 @@ class EntsoePandasClient(EntsoeRawClient):
         df = pd.concat(data.values(), axis=1, keys=data.keys())
         df = df.truncate(before=start, after=end)
         return df
+
